# Creates a grid of matplotlib axes.
# Copyright (C) 2025 Dan Crawford
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

try:
    from matplotlib import colors
    cont = True
except:
    cont = False

if cont:
    import numpy as np
    def build_axes_grid(fig, nx, ny, width=None, height=None, xoff=0.05, yoff=0.05, xspace=None, yspace=None, aspect=None, axes_args=None, share=True):
        """
        Build a grid of matplotlib axes

        Parameters
        ----------
        fig: matplotlib.figure.Figure
        nx, ny: int
            number of axes in the x and y directions
        width, height: float
            width and height of each axis
        xoff, yoff: float | pair of floats
            margin between the boundaries of the figure and the extremal axes.
            If a float, then the same margin is applied to the left/right or top/bottom.
            Otherwise, if a pair than the appropriate values are applied.
        xspace, yspace: float
            Spacing between axes
        aspect: float
            Aspect ratio of the axes
        axes_args:
            additional arguments passed to figure.add_axes
        share: bool
            if the axes should be shared, so that x/y ticklabels are only shown on the left and bottom axes

        Returns
        ----------
        axes: numpy array of matplotlib.axis.Axis
            Note that the origin coresponds to the bottom left axis.
        """
        if isinstance(xoff, int) or isinstance(xoff, float):
            xoff = (xoff, xoff)
        if isinstance(yoff, int) or isinstance(yoff, float):
            yoff = (yoff, yoff)
        if not xspace is not None:
            xspace = 0.1/nx
        if not yspace is not None:
            yspace = 0.1/ny
        if not width is not None:
            width = (1-xoff[0]-xoff[1]-(nx-1)*xspace)/nx
        if not height is not None:
            height = (1-yoff[0]-yoff[1]-(ny-1)*yspace)/ny

        fig_width, fig_height = fig.bbox_inches._points[1]

        if aspect is not None:
            if fig_width < fig_height:
                height = width * aspect * fig_width / fig_height
            else:
                width = height * aspect * fig_height / fig_width

        if nx == ny == 1:
            if axes_args is not None:
                kwargs = axes_args[i, j]
            else:
                kwargs = {}
            i = 0
            j = 0
            axis = fig.add_axes([xoff[0]+i*(xspace+width), yoff[0]+j*(yspace+height), width, height], **kwargs)
            return axis

        axes = np.empty((nx, ny), dtype=object)
        for i, j in np.ndindex(nx, ny):
            if axes_args is not None:
                kwargs = axes_args[i, j]
            else:
                kwargs = {}
            axes[i, j] = fig.add_axes([xoff[0]+i*(xspace+width), yoff[0]+j*(yspace+height), width, height], **kwargs)
        if share:
            for i, j in np.ndindex(nx, ny):
                if i > 0:
                    axes[i, j].sharey(axes[0, j])
                    axes[i, j].tick_params(axis="y", labelleft=False)
                if j > 0:
                    axes[i, j].sharex(axes[i, 0])
                    axes[i, j].tick_params(axis="x", labelbottom=False)

        return np.squeeze(axes) 
else:
    def build_axes_grid(fig, nx, ny, width=None, height=None, xoff=0.05, yoff=0.05, xspace=None, yspace=None, aspect=None, axes_args=None, share=True):
        raise NotImplementedError("matplotlib not installed!")



# Produces a set of points corresponding to a rectangle with rounded corners
# Modified from <https://stackoverflow.com/a/66279687>, user tenhjo
# CC BY-SA 4.0

try:
    from matplotlib.patches import PathPatch
    from matplotlib import path
    cont = True
except:
    cont = False

if cont:
    import numpy as np

    class RoundedPolygon(PathPatch):
        def __init__(self, xy, pad, **kwargs):
            p = path.Path(*self.__round(xy=xy, pad=pad))
            super().__init__(path=p, **kwargs)

        def __round(self, xy, pad):
            n = len(xy)

            for i in range(0, n):

                x0, x1, x2 = np.atleast_1d(xy[i - 1], xy[i], xy[(i + 1) % n])

                d01, d12 = x1 - x0, x2 - x1
                d01, d12 = d01 / np.linalg.norm(d01), d12 / np.linalg.norm(d12)

                x00 = x0 + pad * d01
                x01 = x1 - pad * d01
                x10 = x1 + pad * d12
                x11 = x2 - pad * d12

                if i == 0:
                    verts = [x00, x01, x1, x10]
                else:
                    verts += [x01, x1, x10]
            codes = [path.Path.MOVETO] + n*[path.Path.LINETO, path.Path.CURVE3, path.Path.CURVE3]

            return np.atleast_1d(verts, codes)
else:
    class RoundedPolygon():
        def __init__():
            pass



# Creates a unique hash for a dict
# Copyright (C) 2025 Dan Crawford
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import hashlib

def hash_dict(d):
    """
    Creates a unique hash for an (arbitrarily complicated) dict.
    """
    for k,v in d.items():
        if type(v) == list:
            d[k] = str(v)
        if type(v) == dict:
            d[k] = hash_dict(v)

    h = hashlib.sha224(
        str(
            sorted(
                list(
                    frozenset(d.items())
                )
            )
        ).encode('utf-8')
    ).hexdigest()[:20]
    return h



# Adds a decorator which enables a function to take vector arguments, and runs in parallel with caching.
# Author: Pauli Virtanen, Dan Crawford
# License: GNU Affero General Public License, see AGPL.txt for details

import functools
import joblib
import inspect
from tqdm.auto import tqdm
import numpy as np


class _ParallelSafeMemFunc:
    """
    Function that updates inspect.linecache when called.
    Required to make joblib.Memory work with parallel execution
    with IPython cells.
    """

    def __init__(self, func, mem):
        self._orig_func = func
        self._func = mem.cache(func)
        inspect.linecache.checkcache()
        inspect.getsourcelines(func)
        self._linecache = inspect.linecache.cache

    def __call__(self, *args, **kwargs):
        if self._linecache is not None:
            inspect.linecache.cache = self._linecache
            self._linecache = None

        self._old_checkcache = inspect.linecache.checkcache
        try:
            inspect.linecache.checkcache = lambda fn=None: None
            return self._func(*args, **kwargs)
        finally:
            inspect.linecache.checkcache = self._old_checkcache


def _unarray(x):
    if isinstance(x, np.ndarray):
        return x.item()
    return x

class ProgressParallel(joblib.Parallel):
    # https://stackoverflow.com/a/61027781
    def __init__(self, *args, **kwargs):
        self.total = kwargs.pop("total")
        self._progress = kwargs.pop("progress", False)
        super().__init__(**kwargs)

    def __call__(self, *args, **kwargs):
        if self._progress:
            with tqdm(total=self.total) as self._pbar:
                return joblib.Parallel.__call__(self, *args, **kwargs)
        return joblib.Parallel.__call__(self, *args, **kwargs)

    def print_progress(self):
        if self._progress:
            self._pbar.n = self.n_completed_tasks
            self._pbar.refresh()


def vectorize_parallel(
    func=None,
    backend=None,
    n_jobs=-1,
    returns_object=False,
    batch_size="auto",
    included=None,
    excluded=None,
    noarray=False,
):
    """
    Decorator to vectorize and function to evaluate elements in parallel.
    Uses *joblib* for parallelization.

    Parameters
    ----------
    func : function
    backend : str
        Joblib backend
    n_jobs : int
        Number of jobs, -1 means number of CPUs.
    returns_object : bool
        Put function return values to an object array.
    batch_size : {"auto", int}
    included : tuple
        Function argument names to include in vectorization.
    excluded : tuple
        Function argument names to exclude from vectorization.
    noarray : bool
        Convert array-valued return values to scalars.
    """

    if func is None:

        def deco(func):
            return vectorize_parallel(
                func,
                backend=backend,
                n_jobs=n_jobs,
                returns_object=returns_object,
                batch_size=batch_size,
                included=included,
                excluded=None,
                noarray=noarray,
            )

        return deco

    argspec = inspect.signature(func)

    excluded = list(excluded) if excluded else []
    excluded += ["params", "mem", "parallel", "verbose",  "progress", "refresh_cache"]
    if included is not None:
        excluded += [
            x.name for x in argspec.parameters.values() if x.name not in included
        ]

    arg_names = [
        x.name
        for x in argspec.parameters.values()
        if x.kind
        in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
    ]
    vararg_names = [
        x.name
        for x in argspec.parameters.values()
        if x.kind == inspect.Parameter.VAR_POSITIONAL
    ]
    if vararg_names:
        vararg_name = vararg_names[0]
    else:
        vararg_name = None

    star_kwargs_name = None
    for x in argspec.parameters.values():
        if x.kind == inspect.Parameter.VAR_KEYWORD:
            star_kwargs_name = x.name
            break

    nonvec_arg_names = []
    for name in list(arg_names):
        if name in excluded:
            arg_names.remove(name)
            nonvec_arg_names.append(name)

    if noarray:
        unarray = _unarray
        unarray_many = lambda x: map(_unarray, x)
    else:
        unarray = lambda x: x
        unarray_many = lambda x: x

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        parallel = kwargs.get("parallel", True)
        if "parallel" not in nonvec_arg_names:
            kwargs.pop("parallel", None)

        mem = kwargs.get("mem", None)
        if "mem" not in nonvec_arg_names:
            kwargs.pop("mem", None)

        verbose = kwargs.get("verbose", False)
        if "verbose" not in nonvec_arg_names:
            kwargs.pop("verbose", False)

        progress = kwargs.get("progress", False)
        if "progress" not in nonvec_arg_names:
            kwargs.pop("progress", False)

        refresh_cache = kwargs.get("refresh_cache", False)
        if "refresh_cache" not in nonvec_arg_names:
            kwargs.pop("refresh_cache", None)

        if refresh_cache:
            cache_validation_callback = lambda x: False
        else:
            cache_validation_callback = lambda x: True

        if mem is not None:
            filename = inspect.getsourcefile(func)
            if parallel and filename.startswith("<") and filename.endswith(">"):
                # Lambdas etc. require saving inspect.linecache
                call_func = _ParallelSafeMemFunc(func, mem)
            else:
                call_func = mem.cache(func, cache_validation_callback=cache_validation_callback, verbose=int(verbose))
        else:
            call_func = func

        boundspec = argspec.bind(*args, **kwargs)
        boundspec.apply_defaults()
        callargs = boundspec.arguments
        args = tuple(callargs.pop(name) for name in arg_names)
        if vararg_name:
            args += callargs.pop(vararg_name)

        if star_kwargs_name is not None:
            extra_kwargs = callargs.pop(star_kwargs_name)
            callargs.update(extra_kwargs)

        v = tuple(map(np.asarray, args))
        v_broadcast = np.broadcast_arrays(*v)

        if len(v_broadcast) > 0:
            total = v_broadcast[0].size
            if total == 1:
                parallel = False
        else:
            total = 0
            parallel = False

        if total > 1:
            it = np.nditer(v, ["refs_ok"])

            if parallel:
                delayed = joblib.delayed
            else:
                delayed = lambda x: x

            # Parallelize
            if len(v) == 1:
                jobs = (delayed(call_func)(unarray(w), **callargs) for w in it)
            else:
                jobs = (delayed(call_func)(*unarray_many(w), **callargs) for w in it)

            if parallel:
                results = ProgressParallel(
                    total=total,
                    n_jobs=n_jobs,
                    backend=backend,
                    batch_size=batch_size,
                    progress=progress,
                )(jobs)
            else:
                results = list(jobs)
        else:
            if len(v) == 0:
                results = [call_func(**callargs)]
            else:
                # Single point
                it = np.nditer(v, ["refs_ok"])
                if len(v) == 1:
                    results = [call_func(unarray(w), **callargs) for w in it]
                else:
                    results = [call_func(*unarray_many(w), **callargs) for w in it]

        if not returns_object:
            results = np.array(results)
        else:
            results_arr = np.zeros(len(results), dtype=object)
            results_arr[:] = results
            results = results_arr
            del results_arr
        results_shape = results.shape[1:]

        if results.shape == (1,):
            return results

        if len(v_broadcast) == 0:
            return results[0]

        if results.ndim > 0:
            results = np.rollaxis(results, 0, results.ndim)
        return results.reshape(results_shape + v_broadcast[0].shape)

    return wrapper



# Discretises the Brillouin zone between certain momenta points.
# Copyright (C) 2025 Dan Crawford, Harald Jeschke
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

def kpath(npoints, points, endpoint=True):
    """Discretises the BZ into `npoints` between endpoints  `points`"""
    nsegments = len(points) - 1
    points_per_segment = []
    for i in range(nsegments):
        if i == nsegments - 1:
            points_per_segment.append(
                npoints - int(npoints/nsegments)*(nsegments - 1)
            )
        else:
            points_per_segment.append(int(npoints/nsegments))

    path = []
    endpoint = 1 if endpoint else 0
    for i in range(nsegments):
        if i == nsegments - 1:
            distance = (np.array(points[i+1]) - np.array(points[i])) / (points_per_segment[i] - endpoint)
        else:
            distance = (np.array(points[i+1]) - np.array(points[i])) / (points_per_segment[i])

        for j in range(int(points_per_segment[i])):
            path_point = points[i] + j * distance
            path.append(path_point)
    return path

k_path = kpath



# Similar to numpy.ndindex, but allows specifying both lower and upper points of the iterators.
# Copyright (C) 2025 Dan Crawford
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

def ndindex(*args):
    """
    Similar to numpy.ndindex but both lower and upper bounds can be specified.

    Example:
        ndindex(10, 20, (-3, 15), (-19, 21), 4)

    """
    def product(*iterables):
        pools = [tuple(pool) for pool in iterables]

        result = [[]]
        for pool in pools:
            result = [x+[y] for x in result for y in pool]

        for prod in result:
            yield tuple(prod)

    ranges = []
    for _range in args:
        if hasattr(_range, "__len__"):
            ranges.append(range(_range[0], _range[1]))
        else:
            ranges.append(range(_range))

    return product(*ranges)



# Various colourmaps and related utilities.
# Copyright (C) 2025 Dan Crawford
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

try:
    import numpy as np
    from matplotlib.colors import LinearSegmentedColormap, ListedColormap, to_rgb, rgb_to_hsv, hsv_to_rgb
    cont = True
except:
    cont = False


default = {
        "blue":         "#193ce6",
        "orange":       "#ff8000",
        "green":        "#29a366",
        "red":          "#ff4e33",
        "violet":       "#bb99ff",
        "brown":        "#8e4b25",
        "pink":         "#ff9999",
        "yellow":       "#e5e51a",
        "lightblue":    "#b1dfe5",
        "mediumblue":   "#267399",
        "greyblue":     "#40b1bf",
        "black":        "#000000",
        "white":        "#fffaf0",
        "purple":       "#6619ff",
}
spectral = {
        "red":          "#ff4e33",
        "orange":       "#ff8000",
        "yellow":       "#e5e51a",
        "lightblue":    "#b1dfe5",
        "blue":         "#193ce6",
        "green":        "#29a366",
        "violet":       "#bb99ff",
}

if cont:
    def hsv_to_html(h, s, v):
        """Convert HSV representation to HTML representation"""
        # Convert HSV to RGB (values between 0 and 1)
        r, g, b = hsv_to_rgb((h, s, v))
        
        # Scale RGB to 0–255
        r = int(round(r * 255))
        g = int(round(g * 255))
        b = int(round(b * 255))
        
        # Format as HTML hex color
        return f'#{r:02X}{g:02X}{b:02X}'

    def hue_interp(h1, h2, n):
        """shortest signed angular distance"""
        dh = ((h2 - h1 + 0.5) % 1.0) - 0.5
        return (h1 + np.linspace(0, dh, n)) % 1.0

    def RGB(l):
        """Convert HTML representation to RGB representation"""
        return [to_rgb(c) for c in l]

    def white_gradient(base_colour):
        """Create a gradient from base_colour to white"""
        h, s, v = rgb_to_hsv(to_rgb(base_colour))
        cmap = [hsv_to_html(h, i, v) for i in np.arange(0, 1.1, 0.1)]
        return cmap

    def black_gradient(base_colour):
        """Create a gradient from base_colour to black"""
        h, s, v = rgb_to_hsv(to_rgb(base_colour))
        cmap = [hsv_to_html(h, i, i) for i in np.arange(0, 1.1, 0.1)]
        return cmap

    def cream_gradient(base_colour):
        """Create a gradient from base_colour to cream"""
        h, s, v = rgb_to_hsv(to_rgb(base_colour))
        cmap = [hsv_to_html(h, i, v) for i in np.arange(0, 1.1, 0.1)]
        cmap[0] = default["white"]
        return cmap

    def black_white_gradient(base_colour):
        """Create a gradient from black to base colour to white"""
        h, s, v = rgb_to_hsv(to_rgb(base_colour))
        cmap1 = [hsv_to_html(h, i, i) for i in np.arange(0, 1.1, 0.1)]
        cmap2 = [hsv_to_html(h, i, v) for i in np.arange(0, 1.1, 0.1)]
        cmap = cmap1 + list(reversed(cmap2))
        return cmap

    def black_cream_gradient(base_colour):
        """Create a gradient from black to base colour to cream"""
        h, s, v = rgb_to_hsv(to_rgb(base_colour))
        cmap1 = [hsv_to_html(h, i, i) for i in np.arange(0, 1.1, 0.1)]
        cmap2 = [hsv_to_html(h, i, v) for i in np.arange(0, 1.1, 0.1)]
        cmap2[0] = default["white"]
        cmap = cmap1 + list(reversed(cmap2))
        return cmap

    def saturation_gradient(base_colour):
        """Create a gradient where the saturation varies"""
        h, s, v = rgb_to_hsv(to_rgb(base_colour))
        cmap = [hsv_to_html(h, i, v) for i in np.arange(0.1, 1.1, 0.1)]
        return cmap

    def value_gradient(base_colour):
        """Create a gradient where the value varies"""
        h, s, v = rgb_to_hsv(to_rgb(base_colour))
        cmap = [hsv_to_html(h, s, i) for i in np.arange(0.1, 1.1, 0.1)]
        return cmap

    def saturation_value_gradient(base_colour):
        """Create a gradient where the value varies, and then the saturation"""
        h, s, v = rgb_to_hsv(to_rgb(base_colour))
        cmap1 = [hsv_to_html(h, i, v) for i in np.arange(0.1, 1, 0.1)]
        cmap2 = [hsv_to_html(h, s, i) for i in np.arange(0.1, 1.1, 0.1)]
        cmap = cmap1 + list(reversed(cmap2))
        return cmap

    def twin_gradient(base_colour1, base_colour2):
        """Create a gradient from one colour, through white, to another colour"""
        h1, s1, v1 = rgb_to_hsv(to_rgb(base_colour1))
        h2, s2, v2 = rgb_to_hsv(to_rgb(base_colour2))

        cmap1 = []
        ss = np.linspace(s1, 0, 10)
        vs = np.linspace(v1, 1, 10)
        for i in range(10):
            cmap1.append(hsv_to_html(h1, ss[i], vs[i]))

        cmap2 = []
        ss = np.linspace(0, s2, 10)
        vs = np.linspace(1, v2, 10)
        for i in range(10):
            cmap2.append(hsv_to_html(h2, ss[i], vs[i]))

        cmap = cmap1 + cmap2

        return cmap

    def twin_cream_gradient(base_colour1, base_colour2):
        """Create a gradient from one colour, through white, to another colour"""
        h1, s1, v1 = rgb_to_hsv(to_rgb(base_colour1))
        h2, s2, v2 = rgb_to_hsv(to_rgb(base_colour2))

        cmap1 = []
        ss = np.linspace(s1, 0.1111, 10)
        vs = np.linspace(v1, 1, 10)
        for i in range(10):
            cmap1.append(hsv_to_html(h1, ss[i], vs[i]))

        cmap2 = []
        ss = np.linspace(0.1111, s2, 10)
        vs = np.linspace(1, v2, 10)
        for i in range(10):
            cmap2.append(hsv_to_html(h2, ss[i], vs[i]))

        cmap = cmap1 + cmap2

        return cmap

    def twin_black_white_gradient(base_colour1, base_colour2):
        """Create a gradient from black, to one colour, to another colour, to white"""
        h1, s1, v1 = rgb_to_hsv(to_rgb(base_colour1))
        h2, s2, v2 = rgb_to_hsv(to_rgb(base_colour2))

        cmap1 = []
        ss = np.linspace(0, s1, 10)
        vs = np.linspace(0, v1, 10)
        for i in range(10):
            cmap1.append(hsv_to_html(h1, ss[i], vs[i]))

        cmap2 = []
        hs = hue_interp(h1, h2, 10)
        ss = np.linspace(s1, s2, 10)
        vs = np.linspace(v1, v2, 10)
        for i in range(10):
            cmap2.append(hsv_to_html(hs[i], ss[i], vs[i]))

        cmap3 = []
        ss = np.linspace(s2, 0, 10)
        vs = np.linspace(v2, 1, 10)
        for i in range(10):
            cmap3.append(hsv_to_html(h2, ss[i], vs[i]))

        cmap = cmap1 + cmap2 + cmap3

        return cmap

    def twin_black_cream_gradient(base_colour1, base_colour2):
        """Create a gradient from black, to one colour, to another colour, to white"""
        h1, s1, v1 = rgb_to_hsv(to_rgb(base_colour1))
        h2, s2, v2 = rgb_to_hsv(to_rgb(base_colour2))

        cmap1 = []
        ss = np.linspace(0, s1, 10)
        vs = np.linspace(0, v1, 10)
        for i in range(10):
            cmap1.append(hsv_to_html(h1, ss[i], vs[i]))

        cmap2 = []
        hs = hue_interp(h1, h2, 10)
        ss = np.linspace(s1, s2, 10)
        vs = np.linspace(v1, v2, 10)
        for i in range(10):
            cmap2.append(hsv_to_html(hs[i], ss[i], vs[i]))

        cmap3 = []
        ss = np.linspace(s2, 0.1111, 10)
        vs = np.linspace(v2, 1, 10)
        for i in range(10):
            cmap3.append(hsv_to_html(h2, ss[i], vs[i]))

        cmap = cmap1 + cmap2 + cmap3

        return cmap

    class Cmap():
        def __init__(self, name, listed=False, gradient="black_white", N=1000, is_reversed=False, callback=None):
            """
            Create a colourmap object

            Parameters
            ----------
            name: str
                name of the base colour to use, drawn from the above default dict.
            listed: bool
                If the colourmap is discrete
            gradient: str
                type of gradient to use
            N: int
                number of colours
            is_reversed: bool
                reverse the colourmap
            callback: function | None
                optional function to generate the colourmap from the base colour

            """
            self.listed = listed
            self.N = N
            self.name = name
            if callback is None:
                match gradient:
                    case "cream":
                        self.colours = cream_gradient(default[name])
                    case "white":
                        self.colours = white_gradient(default[name])
                    case "black":
                        self.colours = black_gradient(default[name])
                    case "black_white":
                        self.colours = black_white_gradient(default[name])
                    case "black_cream":
                        self.colours = black_cream_gradient(default[name])
                    case "saturation":
                        self.colours = saturation_gradient(default[name])
                    case "value":
                        self.colours = value_gradient(default[name])
                    case "saturation_value":
                        self.colours = saturation_value_gradient(default[name])
                    case "twin":
                        self.colours = twin_gradient(default[name[0]], default[name[1]])
                    case "twin_cream":
                        self.colours = twin_cream_gradient(default[name[0]], default[name[1]])
                    case "twin_black_white":
                        self.colours = twin_black_white_gradient(default[name[0]], default[name[1]])
                    case "twin_black_cream":
                        self.colours = twin_black_cream_gradient(default[name[0]], default[name[1]])
                    case _:
                        raise ValueError("Invalid gradient")
            else:
                self.colours = callback(default[name])
            if is_reversed and self.colours is not None:
                self.colours = list(reversed(self.colours))

        def __call__(self):
            if self.listed:
                return ListedColormap(RGB(self.colours), name=f"{self.name}s")
            return LinearSegmentedColormap.from_list(f"{self.name}s", RGB(self.colours), self.N)

        def __getitem__(self, index):
            return self.colours[index]

    # Various colourmaps below

    class Blues(Cmap):
        def __init__(self, **kwargs):
            super().__init__("blue", **kwargs)

    class Oranges(Cmap):
        def __init__(self, **kwargs):
            super().__init__("orange", **kwargs)

    class Greens(Cmap):
        def __init__(self, **kwargs):
            if "gradient" in kwargs:
                if kwargs["gradient"] == "special":
                    kwargs["gradient"] = None
                    kwargs["callback"] = self.special_gradient
            super().__init__("green", **kwargs)

        def special_gradient(self, base_colour):
            h, s, v = rgb_to_hsv(to_rgb(base_colour))
            sats = [0.1265, 0.2638, 0.3616, 0.57, 0.75, 0.7484, 0.7459, 0.7561, 0.7561]
            vals = [0.9607, 0.9215, 0.9215, 0.8392, 0.8, 0.6392, 0.4784, 0.3215, 0.1607]
            cmap = [hsv_to_html(h, i, j) for i, j in zip(sats, vals)]
            return cmap

    class Reds(Cmap):
        def __init__(self, **kwargs):
            super().__init__("red", **kwargs)

    class Violets(Cmap):
        def __init__(self, **kwargs):
            super().__init__("violet", **kwargs)

    class LightBlues(Cmap):
        def __init__(self, **kwargs):
            super().__init__("lightblue", **kwargs)

    class MediumBlues(Cmap):
        def __init__(self, **kwargs):
            super().__init__("mediumblue", **kwargs)

    class GreyBlues(Cmap):
        def __init__(self, **kwargs):
            if kwargs["gradient"] == "special":
                kwargs["gradient"] = None
                kwargs["callback"] = self.special_gradient
            super().__init__("greyblue", **kwargs)

        def special_gradient(self, base_colour):
            h, s, v = rgb_to_hsv(to_rgb(base_colour))
            sats = [0.1033, 0.2183, 0.3548, 0.5, 0.6649, 0.6667, 0.6695, 0.6753, 0.6579]
            vals = [0.949, 0.898, 0.8509, 0.8, 0.7490, 0.6, 0.4509, 0.3019, 0.149]
            cmap = [hsv_to_html(h, i, j) for i, j in zip(sats, vals)]
            return cmap

    class Purples(Cmap):
        def __init__(self, **kwargs):
            kwargs["gradient"] = None
            kwargs["callback"] = lambda c: None
            super().__init__("greyblue", **kwargs)
            self.colours = ["#110033", "#220066", "#330099", "#4400cc", "#5500ff", "#7733ff", "#9966ff", "#bb99ff", "#ddccff"]

    class OrangesPurples(Cmap):
        def __init__(self, **kwargs):
            if "gradient" not in kwargs:
                kwargs["gradient"] = "twin_cream"
            super().__init__(["orange", "purple"], **kwargs)

    class PurplesOranges(Cmap):
        def __init__(self, **kwargs):
            if "gradient" not in kwargs:
                kwargs["gradient"] = "twin_cream"
            super().__init__(["purple", "orange"], **kwargs)

    class RedBlue(Cmap):
        def __init__(self, **kwargs):
            if "gradient" not in kwargs:
                kwargs["gradient"] = "twin_cream"
            super().__init__(["red", "blue"], **kwargs)

    class BlueRed(Cmap):
        def __init__(self, **kwargs):
            if "gradient" not in kwargs:
                kwargs["gradient"] = "twin_cream"
            super().__init__(["blue", "red"], **kwargs)

    class GreenPurple(Cmap):
        def __init__(self, **kwargs):
            if "gradient" not in kwargs:
                kwargs["gradient"] = "twin_cream"
            super().__init__(["green", "purple"], **kwargs)

    class PurpleGreen(Cmap):
        def __init__(self, **kwargs):
            if "gradient" not in kwargs:
                kwargs["gradient"] = "twin_cream"
            super().__init__(["purple", "green"], **kwargs)

    class OrangeLightBlue(Cmap):
        def __init__(self, **kwargs):
            if "gradient" not in kwargs:
                kwargs["gradient"] = "twin_cream"
            super().__init__(["orange", "lightblue"], **kwargs)

    class LightBlueOrange(Cmap):
        def __init__(self, **kwargs):
            if "gradient" not in kwargs:
                kwargs["gradient"] = "twin_cream"
            super().__init__(["lightblue", "orange"], **kwargs)

    class GreenMediumBlue(Cmap):
        def __init__(self, **kwargs):
            if "gradient" not in kwargs:
                kwargs["gradient"] = "twin_cream"
            super().__init__(["green", "mediumblue"], **kwargs)

    class MediumBlueGreen(Cmap):
        def __init__(self, **kwargs):
            if "gradient" not in kwargs:
                kwargs["gradient"] = "twin_cream"
            super().__init__(["mediumblue", "green"], **kwargs)

    def Spectral(N=1000, is_reversed=True, lower="default", upper="default"):
        colours = list(spectral.values())
        match upper:
            case "white":
                colours = [default["white"]] + colours
            case "black":
                colours = ["#000000"] + colours
        match lower:
            case "white":
                colours = colours + [default["white"]]
            case "black":
                colours = colours + ["#000000"]
        if is_reversed:
            colours = reversed(colours)
        return LinearSegmentedColormap.from_list("Spectral", RGB(colours), N)

    def BlackWhite(transparent=False):
        if transparent:
            return LinearSegmentedColormap.from_list("BlackWhite", [(0, 0, 0), (1, 1, 1, 0)], 2)
        return LinearSegmentedColormap.from_list("BlackWhite", [(0, 0, 0), (1, 1, 1)], 2)
    def WhiteBlack():
        return LinearSegmentedColormap.from_list("WhiteBlack", [(1, 1, 1), (0, 0, 0)], 2)
    def Pastels(N=1000):
        oranges = Oranges(N=5, gradient="saturation", is_reversed=True)
        oranges = oranges[2::2] + ["#fff3e6"]
        blues = GreyBlues(N=5, gradient="special")
        blues = blues[:5]
        greens = Greens(gradient="special", is_reversed=True)
        greens = greens[4:]
        purples = Purples(N=5)
        purples = purples[::-1][:5]
        return LinearSegmentedColormap.from_list("Pastels", oranges + blues + greens + purples, N)
else:
    class Blues():
        def __init__(self, **kwargs):
            pass

    class Oranges():
        def __init__(self, **kwargs):
            pass

    class Greens():
        def __init__(self, **kwargs):
            pass

    class Reds():
        def __init__(self, **kwargs):
            pass

    class Violets():
        def __init__(self, **kwargs):
            pass

    class LightBlues():
        def __init__(self, **kwargs):
            pass

    class MediumBlues():
        def __init__(self, **kwargs):
            pass

    class GreyBlues():
        def __init__(self, **kwargs):
            pass

    class Purples():
        def __init__(self, **kwargs):
            pass

    class OrangesPurples():
        def __init__(self, **kwargs):
            pass

    class RedBlue():
        def __init__(self, **kwargs):
            pass

    class BlueRed():
        def __init__(self, **kwargs):
            pass

    class GreenPurple():
        def __init__(self, **kwargs):
            pass

    class PurpleGreen():
        def __init__(self, **kwargs):
            pass

    class OrangeLightBlue():
        def __init__(self, **kwargs):
            pass

    class LightBlueOrange():
        def __init__(self, **kwargs):
            pass

    class MediumBlueGreen():
        def __init__(self, **kwargs):
            pass

    class GreenMediumBlue():
        def __init__(self, **kwargs):
            pass

    def Spectral(N=1000):
        pass

    def BlackWhite(transparent=False):
        pass

    def WhiteBlack():
        pass

    def Pastels(N=1000):
        pass


__all__ = ["Cmap", "default",  "RGB", "Blues", "Oranges", "Greens", "Reds", "Violets", "LightBlues", "MediumBlues", "Spectral", "BlackWhite", "WhiteBlack", "RedBlue", "BlueRed", "GreenPurple", "PurpleGreen", "OrangeLightBlue", "LightBlueOrange", "Pastels", "OrangesPurples", "PurplesOranges", "MediumBlueGreen", "GreenMediumBlue"]



# Creates a list of indices between the given points
# Copyright (C) 2025 Dan Crawford
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

def int_path(points, wrap=False):
    """
    Create a list of indices (a path of integers) between
    the given points.
    wrap = True wraps the path on itself, including indices 
    between the last and first point
    """
    path = []
    for (x0, y0), (x1, y1) in zip(points, points[1:]):
        if y0 == y1:
            if x0 > x1:
                for x in range(x0, x1, -1):
                    path.append((x, y0))
            else:
                for x in range(x0, x1):
                    path.append((x, y0))
        if x0 == x1:
            if y0 > y1:
                for y in range(y0, y1, -1):
                    path.append((x0, y))
            else:
                for y in range(y0, y1):
                    path.append((x0, y))
    if wrap:
        x0, y0 = points[-1]
        x1, y0 = points[0]
        if y0 == y1:
            if x0 > x1:
                for x in range(x0, x1, -1):
                    path.append((x, y0))
            else:
                for x in range(x0, x1):
                    path.append((x, y0))
        if x0 == x1:
            if y0 > y1:
                for y in range(y0, y1, -1):
                    path.append((x0, y))
            else:
                for y in range(y0, y1):
                    path.append((x0, y))
        # path.append((x1, y1))
    return path



# Handy class for containing parameters.
# Copyright (C) 2025 Dan Crawford
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections.abc import Mapping
import ast
import copy
import numpy as np
from .hash_dict import hash_dict

class Params(Mapping):
    def __init__(self, params, excluded_keys=[]):
        """
        Converts a dict of parameters into a friendly object.

        Example:
            params = Params(dict(
                # Some parameters for a calculation
                mu = 1,
                Delta = 0.1,
                alpha = 0.5,
                sub = dict(
                    beta = 2,
                    kappa = -0.1
                    ),
                ))

            # can now access elements as attributes
            mu = params.mu
            kappa = params.sub.kappa
            # or as strings (useful for metaprogramming)
            beta = params["sub.beta"]

            # can similarly add elements
            params.foo = "foo"
            params.sub.bar = "bar"

            # and remove items
            params.pop("Delta")

            # unique id
            hash = params.hash()

        Parameters
        ----------
        params: dict | str | numpy.ndarray
            Create a params object from a dict, or a str representation,
            or a numpy-encoded representation. The latter, for example,
            might come from np.savez("data.npz", params=params.items()).
        excluded_keys: list
            All subdictionaries are also converted to params objects,
            unless their key is in this list. Handy for exceptionally
            complicated parameters objects.
        """
        if isinstance(params, str):
            params = self.from_str(params)
        if isinstance(params, np.ndarray):
            params = dict(params.tolist())

        self.excluded_keys = excluded_keys
        for k, v in params.items():
            if type(v) == dict and k not in self.excluded_keys:
                setattr(self, k, Params(v))
            elif type(v) == list:
                setattr(self, k, np.array(v))
            else:
                setattr(self, k, v)

    def from_str(self, params):
        tmp = str(params)
        tmp = tmp[10:len(tmp)-1]
        tmp = tmp.replace("array(", "")
        tmp = tmp.replace("])", "]")
        params = ast.literal_eval(tmp)
        return params

    def __len__(self):
        return len(vars(self))

    def get(self, key, value=None):
        if "." in key:
            key, *rest = key.split(".")
            rest = ".".join(rest)
            if key in vars(self):
                try:
                    return vars(self)[key].get(rest, value)
                except:
                    return vars(self)[key]
        if key in vars(self):
            value = vars(self)[key]
        return value

    def __iter__(self):
        for k in vars(self):
            yield k

    def __getitem__(self, key):
        if "." in key:
            key, *rest = key.split(".")
            rest = ".".join(rest)
            if key in vars(self):
                return vars(self)[key][rest]
        return vars(self)[key]

    def __setitem__(self, key, value):
        if "." in key:
            key, *rest = key.split(".")
            rest = ".".join(rest)
            vars(self)[key][rest] = value
        else:
            if type(value) == dict and key not in self.excluded_keys:
                setattr(self, key, Params(value))
            else:
                setattr(self, key, value)

    def to_dict(self):
        tmp = {}
        for k in vars(self):
            v = vars(self)[k]
            if type(v) == Params:
                v = v.to_dict()
            if type(v) == dict:
                v = v.copy()
            if type(v) == np.ndarray:
                v = v.tolist()
            tmp[k] = v
        return tmp

    def hash(self):
        tmp = self.to_dict()
        return hash_dict(tmp)

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return Params(vars(self))

    def __deepcopy__(self):
        return Params(copy.deepcopy(vars(self)))

    def __repr__(self):
        return f"{vars(self)}"

    def pop(self, key):
        v = self[key]
        delattr(self, key)
        return v

    def pretty(self, units):
        """Pretty-print the parameters"""
        string = "Parameters(\n"
        l = len(vars(self))
        r = len(units)
        assert  l == r, f"{l}, {r}"
        for k,u in zip(vars(self), units):
            u = u if u is not None else ""
            v = self[k]
            string = f"{string}\t{k}: {v}{u}\n"
        string = f"{string})"
        return string.expandtabs(4)



# Prints a matrix nicely in a Jupyter notebook
# Copyright (C) 2025 Dan Crawford
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

try:
    from IPython.display import HTML, display
    cont = True
except:
    cont = False

if cont:
    def pprint(H):
        display(HTML(
           '<table><tr>{}</tr></table>'.format(
               '</tr><tr>'.join(
                   '<td>{}</td>'.format('</td><td>'.join(str(_) for _ in row)) for row in H)
               )
        ))
else:
    def pprint(H):
        raise NotImplementedError("IPython is not installed")



from .hash_dict import hash_dict
from .kpath import kpath, k_path
from .rounded_polygon import RoundedPolygon
from .params import Params
from .pprint_array import pprint
from .build_axes import *
from .versioned_import import versioned_import
from .add_colourbar import add_colourbar
from .copy_artist import copy_artist
from .ndindex import ndindex
from .int_path import int_path
from .vectorize_parallel import vectorize_parallel

from importlib.metadata import version, PackageNotFoundError
try:
    __version__ = version("keino")
except PackageNotFoundError:
    # package is not installed
    pass



# Copy a matplotlib artist
# Copyright (C) 2025 Dan Crawford
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

def copy_artist(from_artist, to_artist, ignore_attributes=[], extra_fields=None):
    """
    Copy a matplotlib artist.
    Currently, artists do not have a built-in method copy(); this function rectifies this oversight.

    Parameters
    ----------
    from_artist: matplotlib.artist.Artist
        artist to copy
    to_artist: matplotlib.artist.Artist
        destination artist. Must be pre-allocated
    ignore_attributes: list[str]
        attributes that should not be copied
    extra_fields: list[str] | None
        additional fields to copy
    """
    # Basic idea from <https://stackoverflow.com/a/59325029>

    def copy_attributes(from_artist, to_artist, attributes):
        for i_attribute  in attributes:
            getattr(to_artist, 
                    'set_' + i_attribute)( getattr(from_artist, 'get_' + i_attribute)() )

    def get_attributes(obj, ignore_attributes):
        obj_methods_list = dir(obj)

        obj_get_attr = []
        obj_set_attr = []
        obj_transf_attr =[]

        for name in obj_methods_list:
            if len(name) > 4:
                prefix = name[0:4]
                if prefix == 'get_':
                    obj_get_attr.append(name[4:])
                elif prefix == 'set_':
                    obj_set_attr.append(name[4:])

        for attribute in obj_set_attr:
            if attribute in obj_get_attr and attribute not in ignore_attributes:
                obj_transf_attr.append(attribute)

        return obj_transf_attr

    attributes = get_attributes(from_artist, ignore_attributes)

    copy_attributes(from_artist, to_artist, attributes)

    if extra_fields is not None:
        for field in extra_fields:
            setattr(to_artist, field, getattr(from_artist, field))



# Creates a matplotlib colourbar.
# Copyright (C) 2025 Dan Crawford
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

try:
    from matplotlib import *
    cont = True
except:
    cont = False

if cont:
    def add_colourbar(fig, axis, im, xoff=0.05, yoff=0, xscale=0.05, yscale=1, horizontal=False, **kwargs):
        """
        Create a matplotlib colourbar

        Parameters
        ----------
        fig: matplotlib.figure.Figure
        axis: matplotlib.axes.Axis
        im: matplotlib.collections.QuadMesh or similar
            matplotlib object which can have a colourbar
        xoff: float
            distance between the right edge of the axis and the left edge of the colourbar
        yoff: float
            distance between the bottom edge of the axis and the bottom edge of the colourbar
        xscale: float
            width of the colourbar = xscale x width of axis
        yscale: float
            height of the colourbar = yscale x height of axis
        horizontal: bool
            if the colourbar should have vertical or horizontal orientation
        kwargs:
            additional arguments to figure.colorbar

        Returns
        ----------
        cb: matplotlib.colorbar.Colorbar
        cax: matplotlib.axes.Axis
            the axis corresponding to the colorbar
        """
        if horizontal:
            xoff = 0
            yoff = 0.05
            xcale = 1
            yscale = 0.05

            pos = axis.get_position()
            cax = fig.add_axes([pos.x0 + xoff, pos.y1 + yoff, xscale * (pos.x1 - pos.x0), yscale * (pos.y1 - pos.y0)])
            cb = fig.colorbar(im, cax=cax, orientation="horizontal", **kwargs)
            cb.ax.xaxis.set_ticks_position("top")
        else:
            pos = axis.get_position()
            cax = fig.add_axes([pos.x1 + xoff, pos.y0 + yoff, xscale * (pos.x1 - pos.x0), yscale * (pos.y1 - pos.y0)])
            cb = fig.colorbar(im, cax=cax, **kwargs)
        return cb, cax
else:
    def add_colourbar(fig, axis, im, xoff=0.05, yoff=0, xscale=0.05, yscale=1, horizontal=False, **kwargs):
        raise NotImplementedError("matplotlib not installed!")
